//! Code to assist with error handling, primarily [MaybeErr].

/// A wrapper around `Option<syn::Error>` which allows errors to be easily combined and created
/// from iterators of [syn::Error].
#[derive(Debug, Default)]
pub(crate) struct MaybeErr(Option<syn::Error>);

impl MaybeErr {
    pub(crate) fn none() -> Self {
        MaybeErr(None)
    }

    pub(crate) fn combine(self, other: Self) -> Self {
        let option = match (self.0, other.0) {
            (Some(left), right) => fold_errs(right, left),
            (left, Some(right)) => fold_errs(left, right),
            _ => None,
        };
        MaybeErr(option)
    }
}

impl From<MaybeErr> for Result<(), syn::Error> {
    fn from(value: MaybeErr) -> Self {
        if let Some(err) = value.0 {
            Err(err)
        } else {
            Ok(())
        }
    }
}

impl From<syn::Error> for MaybeErr {
    fn from(err: syn::Error) -> Self {
        MaybeErr(Some(err))
    }
}

impl FromIterator<syn::Error> for MaybeErr {
    fn from_iter<T: IntoIterator<Item = syn::Error>>(iter: T) -> Self {
        MaybeErr(iter.into_iter().fold(None, fold_errs))
    }
}

fn fold_errs(accum: Option<syn::Error>, curr: syn::Error) -> Option<syn::Error> {
    if let Some(mut accum) = accum {
        accum.combine(curr);
        Some(accum)
    } else {
        Some(curr)
    }
}
